local game = require 'game'
local sprite = game.sprite
local social = game.socialHelper
local self = sprite:create(game.gm.getLocalIndicator(0, "self"),'rank_master')
local analytics = game.analyticsHelper
sprite:clearcache()


--[[
    An implementation of Glicko2 rating system.
    
    See:
        http://www.glicko.net/glicko/glicko2.pdf
        
    This is a "translation" to Lua of:
        https://github.com/sublee/glicko2/blob/master/glicko2.py
        
    Globals:
        Glicko2, Rating, WIN, LOSS
]]

WIN     = 1
LOSS    = 0

local EPSILON = 0.000001
local Q       = math.log(10) / 400

----------------------------------------------------------------
--                    Rating
----------------------------------------------------------------

Rating = class()

function Rating:new(mu, phi, sigma)
    self.mu = mu
    self.phi = phi
    self.sigma = sigma
end

function Rating:toString()
    return "mu="..self.mu.."; phi="..self.phi.."; sigma="..(self.sigma or "?")
end

----------------------------------------------------------------
--                    Glicko2
----------------------------------------------------------------

--[[
    Usage:
        local g2 = Glicko2(1500, 0.2)
        g2.rate(myRating, {{WIN, playerARating},
                           {LOSS, playerBRating},
                           {WIN, playerCRating}})
]]

Glicko2 = class()

function Glicko2:new(zero_mu, tau)
    self.zero_mu = zero_mu
    self.tau = tau
end

function Glicko2:scale_down(obj)
    return Rating((obj.mu - self.zero_mu) / 173.7178, obj.phi / 173.7178, obj.sigma)
end

function Glicko2:scale_up(obj)
    return Rating(obj.mu * 173.7178 + self.zero_mu, obj.phi * 173.7178, obj.sigma)
end

function Glicko2:reduce_impact(rating)
    return 1 / math.sqrt(1 + (3 * rating.phi^2) / (math.pi^2))
end

function Glicko2:expect_score(rating, other_rating, impact)
    return 1 / (1 + math.exp(-impact * (rating.mu - other_rating.mu)))
end

function Glicko2:determine_sigma(rating, difference, variance)
    local phi = rating.phi
    local difference_squared = difference^2
    local alpha = math.log(rating.sigma^2)

    local function f(x)
        local t = phi^2 + variance + math.exp(x)
        local a = math.exp(x) * (difference_squared - t) / (2 * t^2)
        local b = (x - alpha) / self.tau^2
        return a - b
    end

    local a = alpha
    local b = 0
    
    if difference_squared > phi^2 + variance then
        b = math.log(difference_squared - phi^2 - variance)
    else
        local k = 1
        while f(alpha - k * math.sqrt(self.tau^2)) < 0 do
            k = k + 1
        end
        b = alpha - k * math.sqrt(self.tau^2)
    end

    local f_a = f(a)
    local f_b = f(b)
    local f_c = 0
    local c = 0

    while math.abs(b - a) > EPSILON do
        c = a + (a - b) * f_a / (f_b - f_a)
        f_c = f(c)
        if f_c * f_b < 0 then
            a, f_a = b, f_b
        else
            f_a = f_a / 2
        end
        b, f_b = c, f_c
    end
    
    return math.exp(1)^(a / 2)
end

function Glicko2:rate(the_rating, other_players)
    local rating = self:scale_down(the_rating)

    local d_square_inv = 0
    local variance_inv = 0
    local difference = 0
    
    for _, result in ipairs(other_players) do
        local actual_score = result[1]
        local other_rating = self:scale_down(result[2])
        local impact = self:reduce_impact(other_rating)
        local expected_score = self:expect_score(rating, other_rating, impact)
        variance_inv = variance_inv + impact^2 * expected_score * (1 - expected_score)
        difference = difference + impact * (actual_score - expected_score)
        d_square_inv = d_square_inv + (expected_score* (1 - expected_score) * Q^2 * impact^2)
    end

    difference = difference / variance_inv
    local variance = 1 / variance_inv
    local denom = rating.phi^(-2) + d_square_inv
    local mu = rating.mu + Q / denom * (difference / variance_inv)
    local phi = math.sqrt(1 / denom)
    local sigma = self:determine_sigma(rating, difference, variance)

    local phi_star = math.sqrt(phi^2 + sigma^2)

    phi = 1 / math.sqrt(1 / phi_star^2 + 1 / variance)
    mu = rating.mu + phi^2 * (difference / variance)

    return self:scale_up(Rating(mu, phi, sigma))
end






