local game = require 'game'
local sprite = game.sprite
local social = game.socialHelper
local self = sprite:create(game.gm.getLocalIndicator(0, "self"),'bot_master')
local analytics = game.analyticsHelper
sprite:clearcache()


--------------------------------------------------------------------------------
-- AI classes
--------------------------------------------------------------------------------

-- Board -----------------------------------------------------------------------

CellType = {
    EMPTY = 0,
    SHOULD_JUMP = 1,
    SHOULD_ROLL = 2,
    SHOULD_GBOOST = 3,
    
    -- Obstacles
    OBSTACLE = 10,
    GROUND = 11,
    ROPE = 12,
}

-- This class is mutable
Board = class()

function Board:new()
    self.data = {}
end

function Board:put(x, y, cellType)
    local key = self:_getKey(x, y)
    self.data[key] = { x, y, cellType }
end

function Board:get(x, y)
    local key = self:_getKey(x, y)
    
    if self.data[key] ~= nil then
        return self.data[key][3]
    else
        return CellType.EMPTY
    end
end

function Board:each(cellType, func)
    for _, v in pairs(self.data) do
        if v[3] == cellType then
            func(self:cursorAt(v[1], v[2]))
        end
    end
end

function Board:rows(cellType, func)
    self:each(cellType, function(c)
        if c:left():get() == cellType then
            return
        end
        
        local length = 1
        while c:right(length):get() == cellType do
            length = length + 1
        end
        
        func(Rect(c.x, c.y, length, 1))
    end)
end

function Board:cols(cellType, func)
    self:each(cellType, function(c)
        if c:up():get() == cellType then
            return
        end
        
        local length = 1
        while c:down(length):get() == cellType do
            length = length + 1
        end
        
        func(Rect(c.x, c.y, 1, length))
    end)
end

function Board:rects(cellType, func)
    local rows = {}
    self:rows(cellType, function(r) table.insert(rows, r) end)
    
    -- Sort by left, then by top, and then by width
    table.sort(rows, function(a, b)
        if a.left == b.left then            
            if a.top == b.top then
                return a.width < b.width
            else
                return a.top < b.top
            end
        else
            return a.left < b.left
        end
    end)
    
    local i = 1
    while i <= #rows do
        local height = 1
        while (i + height <= #rows) and 
                (rows[i]:offsetBy(0, height) == rows[i + height]) do
            height = height + 1
        end
        
        func(Rect(rows[i].left, rows[i].top, rows[i].width, height))
        
        i = i + height
    end
end

function Board:cursorAt(x, y)
    return Cursor(self, x, y)
end

function Board:_getKey(x, y)
    return x.."-"..y
end

-- Cursor ----------------------------------------------------------------------

Cursor = class()

function Cursor:new(board, x, y, brush)
    self.board = board
    self.x = x
    self.y = y
    self.brush = brush or Brush()
end

function Cursor:withBrush(brush)
    return Cursor(self.board, self.x, self.y, brush)
end

function Cursor:left(n)
    return self:_move(-1, 0, n or 1)
end

function Cursor:right(n)
    return self:_move(1, 0, n or 1)
end

function Cursor:up(n)
    return self:_move(0, -1, n or 1)
end

function Cursor:down(n)
    return self:_move(0, 1, n or 1)
end

function Cursor:get()
    return self.board:get(self.x, self.y)
end

function Cursor:put(cellType)
    self.board:put(self.x, self.y, cellType)
    return self
end

function Cursor:fill()
    self.brush:fill(self)
    return self
end

function Cursor:isObstacle()
    return (self:get() == CellType.OBSTACLE) or (self:get() == CellType.GROUND)
end

function Cursor:isEmpty()
    return self:get() == CellType.EMPTY
end

function Cursor:_move(dx, dy, n)
    if n == 0 then
        return self
    end
    
    local newCursor = Cursor(self.board, self.x + dx, self.y + dy, self.brush)
    
    if newCursor.brush:shouldStop(newCursor) then
        return newCursor
    else
        newCursor:fill()
        return newCursor:_move(dx, dy, n - 1)
    end
end

-- Brush -----------------------------------------------------------------------

Brush = class()

function Brush:new(fillWith, stopIf, canOverride)
    self._fillWith = fillWith
    self._stopIf = stopIf
    self._canOverride = canOverride or false
end

function Brush:fillWith(fillWith)
    return Brush(fillWith, self._stopIf, self._canOverride)
end

function Brush:stopIf(stopIf)
    return Brush(self._fillWith, stopIf, self._canOverride)
end

function Brush:canOverride()
    return Brush(self._fillWith, self._stopIf, true)
end

function Brush:fill(cursor)
    if self._fillWith ~= nil then
        if self._canOverride or cursor:isEmpty() then
            cursor:put(self._fillWith)
        end
    end
end

function Brush:shouldStop(cursor)
    if self._stopIf ~= nil then
        return self._stopIf(cursor)
    else
        return false
    end
end

Brushes = {
    JUMP = Brush():
            fillWith(CellType.SHOULD_JUMP):
            stopIf(function(c) return c:isObstacle() end),
        
    OVERRIDE_WITH_JUMP = Brush():
            fillWith(CellType.SHOULD_JUMP):
            canOverride(),
            
    ROLL = Brush():
            fillWith(CellType.SHOULD_ROLL):
            stopIf(function(c) return c:isObstacle() end),
}

-- Partition -------------------------------------------------------------------

Partition = class()

local RESOLUTION = 48

Selection = {
    EVERY_CELL = 0,
    CENTER_ONLY = 1,
    CENTER_BOTTOM = 2, -- Few pixels below center
}

function Partition:new(partitionSprite)
    self.boundingBox = Rect:fromSprite(partitionSprite)
    self.originX = partitionSprite.XAABB
    self.originY = partitionSprite.YAABB
    self.board = Board()
end

function Partition:isInPartition(sprite)
    return self.boundingBox:containsPoint(
                Rect:fromSprite(sprite):getCenterPoint())
end

function Partition:forEachSpriteInPartition(spriteType, func)
    for i = 1, spriteType.Length do
        if self:isInPartition(spriteType[i]) then
            func(spriteType[i])
            Task:yield()
        end
    end
end

function Partition:toIndices(rect)
    return rect:offsetBy(-self.originX, -self.originY):scaleBy(1 / RESOLUTION)
end

function Partition:fromIndices(rect)
    return rect:scaleBy(RESOLUTION):offsetBy(self.originX, self.originY)
end

function Partition:forEachCellContainingSprite(spriteType, selection, func)
    self:forEachSpriteInPartition(spriteType, function(sprite)
        local spriteBb = Rect:fromSprite(sprite)
        local smallerBb = spriteBb:shrinkBy(13)
        
        if selection == Selection.CENTER_ONLY then
            smallerBb = smallerBb:shrinkBy(1000)
        elseif selection == Selection.CENTER_BOTTOM then
            smallerBb = smallerBb:shrinkBy(1000):offsetBy(0, -spriteBb.height / 2)
        end
        
        local indices = self:toIndices(smallerBb)
        indices:forEachPoint(function(x, y)
            func(self.board:cursorAt(x, y), sprite)
            Task:yield()
        end)
    end)
end

function Partition:forEachCell(cellType, func)
    self.board:each(cellType, function(...)
        func(...)
        Task:yield()
    end)
end

function Partition:createMarkerSprites()
    local function putMarker(factory)
        return function(rect)
            rect = self:fromIndices(rect)
            factory {
                x = rect.left,
                y = rect.top,
                width = rect.width,
                height = rect.height,
            }
            Task:yield()
        end
    end

    self.board:rects(CellType.SHOULD_JUMP, putMarker(bot_master.create_jump_marker))
    self.board:rows(CellType.SHOULD_ROLL, putMarker(bot_master.create_roll_marker))
    
    if debug_show_ai_markers == 1 then
        self.board:rects(CellType.OBSTACLE, putMarker(bot_master.create_ai_marker))
        self.board:rects(CellType.ROPE, putMarker(bot_master.create_ai_marker))
    end
end

--------------------------------------------------------------------------------
-- AI logic
--------------------------------------------------------------------------------

g_aiTaskManager = TaskManager()

local function aiTask(partition)
    print("bot_master: Marker creation started")

    -- Read sprites --------------------------------------------------------
    
    partition:forEachCellContainingSprite(rope_tile, Selection.CENTER_ONLY, function(c)
        c:put(CellType.ROPE)
    end)
    
    partition:forEachCellContainingSprite(obstacle_cap, Selection.EVERY_CELL, function(c)
        c:put(CellType.OBSTACLE)
    end)

    partition:forEachCellContainingSprite(gnd_hb, Selection.EVERY_CELL, function(c)
        c:put(CellType.GROUND)
    end)

    -- For now, upright spikes are considered 1x2 walls.
    -- Selecting only the center of spikes, as their sprite is much larger 
    -- than their hitbox.
    partition:forEachCellContainingSprite(spike, Selection.CENTER_ONLY, function(c, sprite)
        if sprite.YScale < 0 then
            return
        end

        c:put(CellType.OBSTACLE)
        c:up():put(CellType.OBSTACLE)
    end)

    -- Create markers ------------------------------------------------------

    -- Roll under narrow passages
    partition:forEachCell(CellType.OBSTACLE, function(c)
        if not c:down():isObstacle() then
            if c:down(2):isObstacle() then
                c:withBrush(Brushes.ROLL):down():left():right(2)
            elseif c:right():isObstacle() then
                c:down():put(CellType.SHOULD_ROLL)
            end
        end
    end)

    -- Jump over vertical obstacles
    partition:forEachCell(CellType.OBSTACLE, function(c)
        
        if c:up():isObstacle() or 
                c:up():left():isObstacle() or
                c:left():isObstacle() then
            return
        end

        -- We're at a place like:
        -- . .
        -- . x
        
        local cCopy = c
        
        local height = 0
        while c:isObstacle() and not c:left():isObstacle() do
            c = c:down()
            height = height + 1
            
            if height > 5 then
                -- Can't jump this wall
                return
            end
        end
        
        c = cCopy
        
        c:up():withBrush(Brushes.JUMP):left(2)
        for i = 1, height + 1 do
            c:withBrush(Brushes.JUMP):left(height + 1)
            c = c:down()
        end
    end)

    -- Jump over spikes
    partition:forEachCellContainingSprite(
            spike, Selection.CENTER_BOTTOM, function(c, sprite)
            
        if sprite.YScale < 0 then
            --[[
                Draw this:
                    g g x g   
                    g g g g     
                    g g g g
                Where 'x' is the spike, and 'g' is gboost/roll.
            ]]
            
            c:withBrush(Brushes.ROLL):
                right():left(3):down():right(4):down():left(4)
        else
            c:withBrush(Brushes.OVERRIDE_WITH_JUMP):up():left():down(2):left():up()
        end
    end)
    
    -- Jump off edges
    partition:forEachCell(CellType.OBSTACLE, function(c)
        if (not c:left():isObstacle()) or
                (c:right():isObstacle()) or
                (c:down():right():isObstacle()) or
                (c:up():isObstacle()) or
                (c:up():left():isObstacle()) or
                (c:up():right():isObstacle()) then
            return
        end
        
        if c:up(3):right(4):isObstacle() then
            return
        end
        
        -- We're at a place like:
        -- . . .
        -- x x .
        -- ? ? .
        
        -- Create a 3x3 jump marker
        c = c:right():withBrush(Brushes.JUMP)
        c:up(1):right(2)
        c:up(2):right(2)
        c:up(3):right(2)
    end)
    
    -- Jump onto platforms
    partition:forEachCell(CellType.OBSTACLE, function(c)
        if (not c:right():isObstacle()) or
                (c:right():up():isObstacle()) or
                (c:up():isObstacle()) or
                (c:left():up():isObstacle()) or
                (c:left():isObstacle()) or
                (c:left():down():isObstacle()) or
                (c:down():isObstacle()) then
            return
        end
        
        -- We're at a place like:
        -- . . .
        -- . x x
        -- . . ?
        
        c = c:left():withBrush(Brushes.JUMP)
        c:up():left(3)
        c:fill():left(3)
        c:down(1):left(4)
        c:down(2):left(4)
        c:down(3):left(4)
    end)
    
    -- Jump off ropes
    partition:forEachCell(CellType.ROPE, function(c)
        if (c:right():get() == CellType.ROPE) or
                (c:right():up():get() == CellType.ROPE) or
                (c:right():down():get() == CellType.ROPE) then
            return
        end
        
        -- We're at a place like:
        -- ? ? .
        -- ? r .
        -- ? ? .
        
        -- Create a 4x4 jump marker
        c = c:down():withBrush(Brushes.JUMP)
        c:right(3)
        c:up(1):right(3)
        c:up(2):right(3)
        c:up(3):right(3)
    end)
    
    -- Jump onto ropes
    partition:forEachCell(CellType.ROPE, function(c)
        if (c:left():get() == CellType.ROPE) or
                (c:left():up():get() == CellType.ROPE) or
                (c:left():down():get() == CellType.ROPE) then
            return
        end
        
        -- We're at a place like:
        -- . ? ?
        -- . r ?
        -- . ? ?
        
        -- Create a 3x3 jump marker
        c = c:left():withBrush(Brushes.JUMP)
        c:down(1):left(2)
        c:down(2):left(2)
        c:down(3):left(2)
    end)
	
	-- Jump above mud
	partition:forEachCellContainingSprite(mud_gnd_front, Selection.EVERY_CELL, function(c, sprite)
        -- Draw 2x2 jump marker above and to the left of the mud.
		c:withBrush(Brushes.JUMP):fill():up():left():down()
    end)
	
	-- Fall onto ropes
	partition:forEachCell(CellType.ROPE, function(c)
		c:up(2):withBrush(Brushes.ROLL):left():up()
    end)
    
	-- The following actions must be the last ones:
	
    -- Don't add markers where designers already added them
    for _, markerType in pairs { 
                jump_marker,
                roll_marker,
                no_markers_here } do
        
        partition:forEachCellContainingSprite(
                markerType, Selection.EVERY_CELL, function(c)
            c:put(CellType.EMPTY)
        end)
        
    end

    partition:createMarkerSprites()

    print("bot_master: Marker creation finished")
end

function g_startAiMarkerCreationForPartition(partition)
   
    if debug_simulator_build == 0 then
        g_aiTaskManager:queueTask(Task(ai_cpu_throttle, function()
            aiTask(partition)
        end))
        -- as addition info you can pass the xpos of the partition
    else
        -- Run directly when in simulator. This makes error windows show actual
        -- stack trace on errors.
        aiTask(partition)
    end
   
end
