local game = require 'game'
local sprite = game.sprite
local social = game.socialHelper
local self = sprite:create(game.gm.getLocalIndicator(0, "self"),'rank_master')
local analytics = game.analyticsHelper
sprite:clearcache()


-- This script runs only once
if rawget(_G, "utils_script_executed") then
    return
end
utils_script_executed = true

-- Class -----------------------------------------------------------------------

--[[
    A simple class factory:
    
        local Foo = class()
    
    Defininig constructor:
    
        function Foo:new(a)
            self.a = a
        end
    
    Defining methods:
    
        function Foo:doSomething(b)
            self.a = self.a + b
        end
        
        function Foo:__eq(other)
            return self.a == other.a
        end
    
    Using the class:
    
        local myInstance = Foo(5)
        myInstance:doSomething(1)
        assert(myInstance.a == 6)
        assert(myInstance == Foo(5)) -- Invokes __eq
]]

function class()
    local classTable = {}
    classTable.__index = classTable
    
    local function newInstance(classTable, ...)
        local instanceTable = setmetatable({}, classTable)
        classTable.new(instanceTable, ...)
        return instanceTable
    end
    
    setmetatable(classTable, {
        __call = newInstance,
    })
    
    return classTable
end

-- Rect ------------------------------------------------------------------------

--[[
    Rectangle, whose coordinates are integers.
    If non-integer values are given, they will be floored.
]]

Rect = class()

-- Static
function Rect:fromSprite(sprite)
    return 
        Rect(sprite.XAABB, sprite.YAABB, sprite.Width * (sprite.XScale / 100), sprite.Height * (sprite.YScale / 100))
end

function Rect:new(left, top, width, height)
    self.left = math.floor(left)
    self.top = math.floor(top)
    self.width = math.floor(width)
    self.height = math.floor(height)
    self.right = left + width
    self.bottom = top + height
end

function Rect:__eq(other)
    return (self.left == other.left) and
           (self.top == other.top) and
           (self.width == other.width) and
           (self.height == other.height)
end

function Rect:shrinkBy(n)
    return Rect(
            min(self.left + n, self.left + self.width / 2),
            min(self.top + n, self.top + self.height / 2),
            max(self.width - n * 2, 0),
            max(self.height - n * 2, 0))
end

function Rect:offsetBy(x, y)
    return Rect(self.left + x, self.top + y, self.width, self.height)
end

function Rect:scaleBy(n)
    return Rect(self.left * n,
                self.top * n,
                self.width * n,
                self.height * n)
end

function Rect:containsPoint(x, y)
    if (x < self.left) or (x > self.right) then
        return false
    end
    
    if (y < self.top) or (y > self.bottom) then
        return false
    end
    
    return true
end

function Rect:getCenterPoint()
    return self.left + self.width / 2, self.top + self.height / 2
end

function Rect:forEachPoint(func)
    for i = self.left, self.right do
        for j = self.top, self.bottom do
            func(i, j)
        end
    end
end

function Rect:toString()
    return "Rect("..self.left..", "..self.top..", "..
                    self.width..", "..self.height..")"
end

-- Task ------------------------------------------------------------------------

Task = class()

local currentTask = nil

-- Static
function Task:yield()
    if currentTask ~= nil then
        currentTask:_yield()
    end
end

function Task:new(yieldRate, func)
    self._co = coroutine.create(func)
    self._yieldCount = 0
    self._yieldRate = yieldRate
end

function Task:isDone()
    return coroutine.status(self._co) == "dead"
end

function Task:continue()
    assert(currentTask == nil)
    assert(self._yieldCount == 0)
    
    currentTask = self
    local success, msg = coroutine.resume(self._co)
    currentTask = nil
    
    if not success then
        error(tostring(msg))
    end
end

function Task:_yield()
    assert(currentTask == self)
    
    self._yieldCount = self._yieldCount + 1
    if self._yieldCount >= self._yieldRate then
        self._yieldCount = 0
        coroutine.yield()
    end
end

TaskManager = class()

function TaskManager:new()
    self._tasks = {}
end

function TaskManager:queueTask(task)
    table.insert(self._tasks, task)
end

function TaskManager:runTask()
    if #self._tasks > 0 then
        print("TaskManager: Continuing last task")
        self._tasks[1]:continue()
        if self._tasks[1]:isDone() then
            print("TaskManager: Removing done task")
            table.remove(self._tasks, 1)
        end
    end
end

-- Set -------------------------------------------------------------------------

Set = class()

function Set:new(list)
    self._items = {}

    for _, v in ipairs(list or {}) do
        self._items[v] = true
    end
end

function Set:add(value)
    self._items[value] = true
end

function Set:remove(value)
    self._items[value] = nil
end

function Set:contains(value)
    return self._items[value] ~= nil
end

function Set:each(func)
    for v, _ in pairs(self._items) do
        func(v)
    end
end


-- Union
function Set:__add(other)
    local result = Set()
    self:each(function(v) result:add(v) end)
    other:each(function(v) result:add(v) end)
    return result
end

-- Substraction
function Set:__sub(other)
    local result = Set()
    self:each(function(v) result:add(v) end)
    other:each(function(v) result:remove(v) end)
    return result
end

-- Intersection
function Set:__mul(other)
    local result = Set()
    self:each(function(v)
        if other:contains(v) then
            result:add(v)
        end
    end)
    return result
end

-- Table utilities -------------------------------------------------------------

function forEachSprite(spriteType, func)
    for i = 1, spriteType.Length do
        func(spriteType[i])
    end
end

function spriteList(spriteType)
    local result = {}
    
    forEachSprite(spriteType, function(sp)
        table.insert(result, sp)
    end)
    
    return result
end

function filter(ls, predicate)
    local result = {}
    for _, v in ipairs(ls) do
        if predicate(v) then
            table.insert(result, v)
        end
    end
    return result
end

function map(ls, func)
    local result = {}
    for k, v in pairs(ls) do
        result[k] = func(v)
    end
    return result
end

--print ------------------------------
function g_TableToString(tbl) 
    local txt = "{"  
    for k,v in pairs(tbl) do        
        txt = txt .. "[" .. k .. "]=" .. v .. ";"
    end
    txt = txt .. "}"    
    return txt
end
--------------------------------------

